
--
--  Macro Broker
--    by Tuhljin
--

MacroBroker = {}
local MacroBroker = MacroBroker

local L = MACROBROKER_STRINGS
local THIS_TITLE = GetAddOnMetadata("MacroBroker", "Title")
local THIS_VERSION = GetAddOnMetadata("MacroBroker", "Version")

local MAX_CHARACTERS_PER_LINE = 255
local QuestionMarkIcon = "Interface\\Icons\\INV_Misc_QuestionMark"

local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local OBJINFO, SAVEDOBJ = {}, {}
local tipOwner, tipObj
local ConditionsMonitorTimer, ConditionsMonitorFrame, objOnEnter = 0

local function chatprint(msg, mode)
  local premsg = mode == 1 and "|cffff4026["..THIS_TITLE.."] |cffff6953" or mode == 2 and "   " or
        "|cffff4026["..THIS_TITLE.."]|r "
  DEFAULT_CHAT_FRAME:AddMessage(premsg..msg);
end


-- CODE SNIPPETS
------------------

local function createSnippetFunc(snippet)
  local func, err = loadstring("return function(self, button, dataobj) "..snippet.." end")
  if (not func) then
    chatprint(L.SNIPPET_FAIL_SYNTAX, 1)
    chatprint(err, 2)
    return;
  end
  return func()
end

local function callSnippetFunc(func, ...)
  local success, ret1 = pcall(func, ...);
  if (not success) then
    chatprint(L.SNIPPET_FAIL_EXECUTION, 1)
    chatprint(ret1, 2)
  end
end


-- RUN MACRO
--------------

local function RunCommands(...)
  local line, slash, success, reason
  for i = 1, select("#", ...) do
    line = select(i, ...)
    slash = strmatch(line, "^(/[^%s]+)")
    if (slash) then  slash = strlower(slash);  end
    if (strsub(line, 1, 1) == "#") then
      -- Ignore comment line
    elseif (slash == "/stopmacro") then  -- Handle special case:
      if (SecureCmdOptionParse(line)) then  return;  end
    else
      if (strlen(line) > MAX_CHARACTERS_PER_LINE) then
        chatprint(L.MACRO_WARN_LINELENGTH, 1)
        line = strsub(line, 1, MAX_CHARACTERS_PER_LINE)
      end
      success, reason = TjRunSlash(line)
      if (not success and reason == "secure") then
        chatprint(L.MACRO_FAIL_SECURE:format(slash), 1)
        return;
      end
    end
  end
end


-- CONDITIONS MONITOR
-----------------------

local function findDynamicIcon(getLink, ...)
  local line = select(1, ...)
  local showtext
  if (strsub(line, 1, 6) == "#show ") then
    showtext = strsub(line, 7)
  elseif (strsub(line, 1, 13) == "#showtooltip ") then
    showtext = strsub(line, 14)
  end
  if (showtext) then
    showtext = strtrim(showtext)
    if (showtext ~= "") then
      showtext = SecureCmdOptionParse(showtext)
      if (showtext and showtext ~= "") then
        local num = tonumber(showtext)
        if (getLink) then
          if (num) then  return GetInventoryItemLink("player", num);  end
          return GetSpellLink(showtext) or select(2,GetItemInfo(showtext))
        end
        return num and GetInventoryItemTexture("player", num) or GetSpellTexture(showtext) or GetItemIcon(showtext) or nil
      end
      return nil;
    end
  end

  local slash, arg
  for i = 1, select("#", ...) do
    line = select(i, ...)
    if (strsub(line, 1, 1) == "#") then
      -- Ignore comment line
    else
      slash, arg = strmatch(line, "^(/[^%s]+)[%s]*(.*)")
      if (slash and arg and arg ~= "") then
        slash = strlower(slash)
        if (slash == "/cast" or slash == "/use") then
          arg = SecureCmdOptionParse(arg)
          if (arg) then
            local num = tonumber(arg)
            if (getLink) then
              if (num) then  return GetInventoryItemLink("player", num);  end
              return GetSpellLink(arg) or select(2,GetItemInfo(arg))
            end
            return num and GetInventoryItemTexture("player", num) or GetSpellTexture(arg) or GetItemIcon(arg) or nil
          end
        end
      end
    end
  end
end

SecureStateDriverManager:HookScript("OnEvent", function()
  ConditionsMonitorTimer = 0
end);

ConditionsMonitorFrame = CreateFrame("Frame")
ConditionsMonitorFrame:Hide()

ConditionsMonitorFrame:SetScript("OnUpdate", function(self, elapsed)
  ConditionsMonitorTimer = ConditionsMonitorTimer - elapsed
  if (ConditionsMonitorTimer <= 0) then
    ConditionsMonitorTimer = STATE_DRIVER_UPDATE_THROTTLE or 0.2
    local keepshown
    for dataobj, info in pairs(OBJINFO) do
      if (info.dynamicIcon) then
        local macro = info.OnClick_macro
        if (macro) then
          keepshown = true
          local icon = findDynamicIcon(false, strsplit("\n", macro)) or QuestionMarkIcon
          if (dataobj.icon ~= icon) then
            dataobj.icon = icon
            if (tipOwner) then  objOnEnter(tipOwner, tipObj);  end
          end
        end
      end
    end
    if (not keepshown) then  self:Hide();  end
  end
end);


-- DATA OBJECT HANDLERS
-------------------------

local function objOnClick(self, button, dataobj)
  local info = OBJINFO[dataobj]
  if (info.OnClick_macro) then
    if (MacroBrokerGUI) then  -- Save macros in case this one has changed
      MacroBrokerGUI.SaveMacro()
    end
    RunCommands( strsplit("\n", info.OnClick_macro) )
  end
  local func = info.OnClick
  if (func) then  callSnippetFunc(func, self, button, dataobj);  end
end

local function objOnTooltipShow(tooltip, dataobj)
  local info = OBJINFO[dataobj]
  tooltip:AddLine(info.name, 1, 1, 1)
  
  local macro = info.OnClick_macro
  if (macro) then
    local line = strlower( strsplit("\n", macro) )
    if (strsub(line, 1, 12) == "#showtooltip") then
      local x = strsub(line, 13, 13)
      if (x == "" or x == " ") then
        local link = findDynamicIcon(true, strsplit("\n", macro))
        if (link) then  tooltip:SetHyperlink(link);  end
      end
    end
  end

  if (info.tiptext) then
    tooltip:AddLine(info.tiptext)
  end
  --[[
  if (not info.notip_footer) then
    tooltip:AddLine(" ")
    tooltip:AddLine("|cff68ccef"..L.CREATEDBYTAG)
    --tooltip:AddDoubleLine("|cff68ccefdynamicIcon:",(info.dynamicIcon and "Y" or "N"))
    --tooltip:AddDoubleLine("|cff68cceficon:",dataobj.icon)
    --if (SAVEDOBJ[dataobj]) then  tooltip:AddDoubleLine("|cff68ccefsaved icon:",SAVEDOBJ[dataobj].icon);  end
  end
  --]]
  if (info.unsaved) then
    tooltip:AddLine(L.PLUGINUNSAVED, 1, 0, 0)
  end
end

objOnEnter = function(self, dataobj)
  if (OBJINFO[dataobj].notip) then  return;  end
  local leftright, topbtm, topbtm2 = ""
  local x, y = self:GetCenter();
  local w, h = UIParent:GetWidth() / 3, UIParent:GetHeight() / 2
  if (x < w) then
    leftright = "LEFT"
  elseif (x > w * 2) then
    leftright = "RIGHT"
  end
  if (y < h) then
    topbtm, topbtm2 = "BOTTOM", "TOP";
  else
    topbtm, topbtm2 = "TOP", "BOTTOM";
  end
  GameTooltip:SetOwner(self, "ANCHOR_NONE")
  GameTooltip:SetPoint(topbtm..leftright, self, topbtm2..leftright)
  objOnTooltipShow(GameTooltip, dataobj)
  GameTooltip:Show()
  tipOwner, tipObj = self, dataobj
end

local function objOnLeave()
  GameTooltip:Hide()
  tipOwner, tipObj = nil, nil
end

local function CreateDataObj(name, t, text, label, icon, value, suffix, tiptext, useOnEnter)
-- Set useOnEnter to true to allow dynamic tooltip updates (changes to the tooltip as it is displayed).
-- Set it to false or nil to allow the LDB display addon to handle tooltips.
  assert(name, "Expected string for first argument (name); got "..type(name))
  t = t or text and "data source" or "launcher"
  if (not text and t == "data source") then  text = "";  end
  local dynamicIcon
  if (icon == "?") then
    icon, dynamicIcon = QuestionMarkIcon, true
  else
    icon = (icon == "" or (not icon and t == "launcher")) and QuestionMarkIcon or icon
  end
  local dataobj = ldb:NewDataObject(name, {
	type = t,
	text = text,
	icon = icon,
	label = label,
	value = value,
	suffix = suffix,
	MacroBroker = true
  })
  if (dataobj) then
    OBJINFO[dataobj] = {
	name = name,
	tiptext = tiptext,
	dynamicIcon = dynamicIcon
    };
    dataobj.OnClick = function(self, button)  objOnClick(self, button, dataobj);  end
    if (useOnEnter) then
      dataobj.OnEnter = function(self)  objOnEnter(self, dataobj);  end
      dataobj.OnLeave = objOnLeave
    else
      dataobj.OnTooltipShow = function(tooltip)  objOnTooltipShow(tooltip, dataobj);  end
    end
    return dataobj, t
  end
end

local function getObjInfo(dataobj)
  if (type(dataobj) == "string") then  dataobj = ldb:GetDataObjectByName(dataobj);  end
  return dataobj and OBJINFO[dataobj], dataobj
end


-- GLOBAL FUNCTIONS
---------------------

-- Note: MacroBroker.CreateDataObj() creates a data object that does not persist across sessions, whereas
-- data objects created by MacroBroker.CreatePlugin() do persist.

MacroBroker.CreateDataObj = CreateDataObj

function MacroBroker.CreatePlugin(name, t, text, label, icon, value, suffix, tiptext, useOnEnter)
  local obj
  obj, t = CreateDataObj(name, t, text, label, icon, value, suffix, tiptext, useOnEnter)
  if (not t) then  return false;  end

  SAVEDOBJ[name] = SAVEDOBJ[name] or {}
  local tab = SAVEDOBJ[name]
  tab.type = t
  tab.text = text
  tab.label = label
  tab.icon = icon
  tab.value = value
  tab.suffix = suffix
  tab.tiptext = tiptext
  tab.useOnEnter = useOnEnter

  return obj
end

function MacroBroker.EditPlugin(dataobj, t, text, label, icon, value, suffix, tiptext, useOnEnter)
-- Use this instead of directly changing the data object's fields when the change should be stored in the
-- saved variables or when changing something handled by Macro Broker that isn't stored in the data object table.
-- NOTE: If an argument is nil or false, the current value is not changed, except for useOnEnter which
-- accepts false (but not nil) as a valid argument.
  local info
  info, dataobj = getObjInfo(dataobj)
  assert(info, "Plugin not found.")
  local saved = SAVEDOBJ[info.name]

  if (text) then  dataobj.text, saved.text = text, text;  end
  if (label) then  dataobj.label, saved.label = label, label;  end
  if (icon) then
    if (icon == "?") then
      dataobj.icon, saved.icon, info.dynamicIcon = QuestionMarkIcon, icon, true
      ConditionsMonitorTimer = 0
      ConditionsMonitorFrame:Show()
    else
      dataobj.icon, saved.icon, info.dynamicIcon = icon, icon, nil
    end
  end
  if (value) then  dataobj.value, saved.value = value, value;  end
  if (suffix) then  dataobj.suffix, saved.suffix = suffix, suffix;  end
  if (tiptext) then  info.tiptext, saved.tiptext = tiptext, tiptext;  end
  -- Your LDB display addon might not recognize this change until it is reloaded:
  if (t) then  dataobj.type, saved.type = t, t;  end
  -- This one won't take effect until the interface is reloaded:
  if (useOnEnter ~= nil) then  saved.useOnEnter = useOnEnter;  end
end

function MacroBroker.EditPluginField(dataobj, field, value)
-- When you need to set a field to nil, use this function instead of EditPlugin. It's also useful when the name
-- of the field is itself a variable.
  local info
  info, dataobj = getObjInfo(dataobj)
  assert(info, "Plugin not found.")
  if (field == "text" and not value and dataobj.type == "data source") then  value = "";  end

  if (field == "type" or field == "text" or field == "label" or field == "value" or field == "suffix") then
    dataobj[field], SAVEDOBJ[info.name][field] = value, value
  elseif (field == "icon") then
    if (value == "?") then
      dataobj.icon, SAVEDOBJ[info.name].icon, info.dynamicIcon = QuestionMarkIcon, value, true
      ConditionsMonitorTimer = 0
      ConditionsMonitorFrame:Show()
    else
      dataobj.icon, SAVEDOBJ[info.name].icon, info.dynamicIcon = value, value, nil
    end
  elseif (field == "tiptext") then
    info[field], SAVEDOBJ[info.name][field] = value, value
  elseif (field == "useOnEnter") then
  -- This one won't take effect until the interface is reloaded:
    SAVEDOBJ[info.name][field] = value
  else
    return false
  end
  return true
end

function MacroBroker.RemovePlugin(dataobj)
-- NOTE: The data object will not actually be removed until the interface is reloaded.
  local info = getObjInfo(dataobj)
  if (info) then
    SAVEDOBJ[info.name].DoNotSave = true
    info.unsaved = true
    return true
  end
end

function MacroBroker.SavePlugin(dataobj)
-- This is used to undo a MacroBroker.RemovePlugin(dataobj) call.
  local info = getObjInfo(dataobj)
  if (info) then
    SAVEDOBJ[info.name].DoNotSave = nil
    info.unsaved = nil
    return true
  end
end

function MacroBroker.IsPluginSaved(dataobj)
  local info = getObjInfo(dataobj)
  if (info and not info.unsaved) then  return true;  end
end

function MacroBroker.SetOnClickSnippet(dataobj, snippet, noSave)
  local info = getObjInfo(dataobj)
  assert(info, "Plugin not found.")
  local func = createSnippetFunc(snippet)
  if (func) then
    info.OnClick = func;
    if (not noSave and SAVEDOBJ[info.name]) then  SAVEDOBJ[info.name].OnClick_snippet = snippet;  end
    return true
  end
end

function MacroBroker.SetOnClickMacro(dataobj, macro, noSave)
  local info = getObjInfo(dataobj)
  assert(info, "Plugin not found.")
  info.OnClick_macro = macro
  if (not noSave and SAVEDOBJ[info.name]) then  SAVEDOBJ[info.name].OnClick_macro = macro;  end
  if (info.dynamicIcon) then
    ConditionsMonitorTimer = 0
    ConditionsMonitorFrame:Show()
  end
  return true
end

function MacroBroker.SetCharacterSpecific(dataobj, perChar, noSave)
  local info = getObjInfo(dataobj)
  assert(info, "Plugin not found.")
  perChar = not not perChar
  info.CharacterSpecific = perChar
  if (not noSave and SAVEDOBJ[info.name]) then  SAVEDOBJ[info.name].CharacterSpecific = perChar;  end
end

function MacroBroker.GetMacroInfo(dataobj)
  local info
  info, dataobj = getObjInfo(dataobj)
  if (info) then
    return dataobj.icon, info.OnClick_macro, info.CharacterSpecific, info.dynamicIcon
  end
end

function MacroBroker.GetRealName(name)
  if (SAVEDOBJ[name]) then  return SAVEDOBJ[name].RealName or name;  end
end

function MacroBroker.SetRealName(name, realname)
  local saved = SAVEDOBJ[name]
  if (saved) then
    if (name == realname) then
      saved.RealName = nil
    else
      saved.RealName = realname
    end
    return true
  end
end

function MacroBroker.IsRealNameUsed(name, exclude)
  for n, saved in pairs(SAVEDOBJ) do
    if (n ~= exclude) then
      if (saved.RealName) then
        if (saved.RealName == name) then  return true;  end
      elseif (n == name) then
        return true
      end
    end
  end
end


-- ITERATOR
-------------

local function ObjectNameIterator(_, i)
  local dataobj, info = next(OBJINFO, i)
  return dataobj, info and info.name
end

function MacroBroker.IterateObjects()
  return ObjectNameIterator, nil
end


-- INITIALIZATION
-------------------

local function RecreateSavedPlugins(vars, vars2, perChar)
  for name,tab in pairs(vars) do
    local obj = MacroBroker.CreatePlugin(name, tab.type, tab.text, tab.label, tab.icon, tab.value, tab.suffix, tab.tiptext, tab.useOnEnter)

    if (not obj) then
      chatprint(L.ERR_OBJNOTCREATED_NAMEUSED:format(name), 1)
      local basename = name.."*"
      local tempname, count = basename, 1
      while (vars[tempname] or (vars2 and vars2[tempname]) or ldb:GetDataObjectByName(tempname)) do
        tempname = basename..count
        count = count + 1
        if (count > 99) then
          chatprint("Error: Could not create plugin \""..name.."\". Data may have been lost.", 1)
        end
      end

      obj = MacroBroker.CreatePlugin(tempname, tab.type, tab.text, tab.label, tab.icon, tab.value, tab.suffix, tab.tiptext, tab.useOnEnter)
      SAVEDOBJ[tempname].RealName = name
      chatprint(L.WARN_OBJUSINGTEMPNAME:format(tempname))
    end

    if (obj) then
      if (perChar) then  MacroBroker.SetCharacterSpecific(obj, true);  end
      if (tab.OnClick_snippet) then
        MacroBroker.SetOnClickSnippet(obj, tab.OnClick_snippet)
      end
      if (tab.OnClick_macro) then
        MacroBroker.SetOnClickMacro(obj, tab.OnClick_macro)
      end
    end
  end
end

local function UpdateDataObjVars(vars, oldver)
  oldver = tonumber(oldver)
  if (not oldver) then  return;  end
  if (oldver < 0.14) then
    for name,tab in pairs(vars) do
      tab.useOnEnter = true
    end
  end
end

local function OnEvent(self, event, arg1)
  if (event == "ADDON_LOADED" and arg1 == "MacroBroker") then
    self:UnregisterEvent("ADDON_LOADED")
    MacroBroker_Vars, MacroBroker_CharVars = MacroBroker_Vars or {}, MacroBroker_CharVars or {}

    if (MacroBroker_Vars.DataObjs) then
      if (THIS_VERSION ~= MacroBroker_Vars.Version) then
        UpdateDataObjVars(MacroBroker_Vars.DataObjs, MacroBroker_Vars.Version)
      end
      RecreateSavedPlugins(MacroBroker_Vars.DataObjs, MacroBroker_CharVars.DataObjs)
      -- MacroBroker_Vars.DataObjs serves no further purpose until it's recreated at logout but we won't remove
      -- it here in case something goes wrong so saved variables aren't lost.
    end
    if (MacroBroker_CharVars.DataObjs) then
      if (THIS_VERSION ~= MacroBroker_CharVars.Version) then
        UpdateDataObjVars(MacroBroker_CharVars.DataObjs, MacroBroker_CharVars.Version)
      end
      RecreateSavedPlugins(MacroBroker_CharVars.DataObjs, nil, true)
      -- MacroBroker_CharVars.DataObjs serves no further purpose until it's recreated at logout but we won't remove
      -- it here in case something goes wrong so saved variables aren't lost.
    end
    UpdateDataObjVars = nil

    MacroBroker_Vars.Version, MacroBroker_CharVars.Version = THIS_VERSION, THIS_VERSION

  elseif (event == "PLAYER_LOGOUT") then
    self:UnregisterEvent("PLAYER_LOGOUT")
    local CHAROBJ
    for name,tab in pairs(SAVEDOBJ) do
      if (tab.DoNotSave) then  -- Remove entries we're not supposed to save
        SAVEDOBJ[name] = nil
      else
        if (tab.CharacterSpecific) then
          -- Move this to the character-specific table
          CHAROBJ = CHAROBJ or {}
          CHAROBJ[tab.RealName or name] = tab
          tab.RealName = nil
          SAVEDOBJ[name] = nil
          tab.CharacterSpecific = nil
        elseif (tab.RealName) then
          SAVEDOBJ[tab.RealName] = tab
          tab.RealName = nil
          SAVEDOBJ[name] = nil
        end
      end
    end

    MacroBroker_CharVars.DataObjs = CHAROBJ
    for _ in pairs(SAVEDOBJ) do
      -- If any entries remain, save them:
      MacroBroker_Vars.DataObjs = SAVEDOBJ
      return;
    end
  end
end

MacroBroker.MainFrame = CreateFrame("Frame")
MacroBroker.MainFrame:SetScript("OnEvent", OnEvent)
MacroBroker.MainFrame:RegisterEvent("ADDON_LOADED")
MacroBroker.MainFrame:RegisterEvent("PLAYER_LOGOUT")


-- SLASH COMMANDS
-------------------

SLASH_MacroBroker1 = "/macrobroker";
SLASH_MacroBroker2 = "/broker";
SLASH_MacroBroker3 = "/mb";
SlashCmdList["MacroBroker"] = function()
  if (not MacroBrokerGUI) then
    local loaded = LoadAddOn("MacroBrokerGUI")
    if (not loaded) then  return;  end
  end
  MacroBrokerGUI.Show()
end


--[[ Some basic test code:
/run MacroBroker.CreatePlugin("Testing")
/run MacroBroker.CreatePlugin("Cool Beans")
/run MacroBroker.RemovePlugin("Testing")
/run MacroBroker.RemovePlugin("Cool Beans")

Snippets - Functional example:
/run MacroBroker.SetOnClickSnippet("Testing", 'print("Testing clicked:", button)')

Snippets - Syntax error test:
/run MacroBroker.SetOnClickSnippet("Cool Beans", 'bad code ahoy ! --print("Testing clicked:", button)')

Snippets - Runtime error test:
/run MacroBroker.SetOnClickSnippet("Cool Beans", 'print("Testing clicked:"..anilvalue, button)')
--]]
