
--
--  eval.lua  -  Simple conditional evaluator for slash commands and chat
--    by Tuhljin
--
--  Implements the /eval command, which allows the use of conditionals to determine the use of other commands or
--  chat. For example, the /calendar command doesn't use conditionals, so "/calendar [mod]" would show the calendar
--  whether a modifier is held or not. With /eval, you can use a command like "/eval [mod] /calendar" to only
--  execute "/calendar" when the conditional evaluation indicates it should.
--
--  IMPORTANT: Secure commands such as /cast, /target, and /focus cannot be used with /eval.
--
--  /eval can has be abbreviated as /ev.
--
--  More examples:
--
--  /eval [flying] I'm flying right now. [mod] I'm not flying.
--    Result: If you are flying, say "I'm flying right now." in your current sticky chat channel (SAY, GUILD, etc.).
--    If not, then if you are holding a modifier key, say "I'm not flying."
--
--  /eval [button:2] /run ShowHelm(not ShowingHelm())
--    Result: If called through a right-click, toggle showing your character's helm.
--
--
--  This file requires the TjRunSlash library.
--


local THIS_VERSION = 0.1

if (not SlashEval or SlashEval.Version < THIS_VERSION) then
  SlashEval = SlashEval or {}
  local oldver = SlashEval.Version or 0
  SlashEval.Version = THIS_VERSION

  local EVAL_FAIL_SECURE = "Error: /eval cannot execute secure commands.";

  if (oldver < 0.1) then
    SLASH_eval1 = "/eval";
    SLASH_eval2 = "/ev";
    SlashCmdList["eval"] = function(...)
      SlashEval.eval(...)
      -- Done this way instead of directly using the code there here because once a given slash command is used,
      -- you can't use 'SlashCmdList["eval"] = <whatever>' to make it point elsewhere. E.g., if the player used
      -- /eval already, you could use such a command to change what /ev calls, but /eval would point to the old
      -- function. This way is more future safe.
    end
  end

  function SlashEval.eval(msg)
    if (msg and msg ~= "") then
      local cmd = SecureCmdOptionParse(msg)
      if (cmd) then
        local success, reason = TjRunSlash(cmd)
        if (not success and reason == "secure") then
          DEFAULT_CHAT_FRAME:AddMessage(EVAL_FAIL_SECURE, 1, 0.15, 0.15)
        end
      end
    end
  end

end
