
--
--  TjRunSlash.lua
--    by Tuhljin
--
--  This library should be accompanied by TjRunSlash.xml, which is the file to include in your TOC or XML file.
--
--  Library for passing slash commands from Lua script to WoW for execution. Any sort of slash commands (from
--  addon-provided commands to emotes) except for secure commands (like /target) can be executed. Non-slash command
--  text can also be handled: It is treated like any other text placed in WoW's primary chat edit box, so it
--  will likely end up in chat.
--
--  Note that while /cast and /use are labeled as secure commands, you are free to use it to open tradeskill windows
--  (through, e.g., "/cast blacksmithing"). (As of version 0.33, TjRunSlash permits the use of conditionals such as
--  "[modifier]") when checking for arguments to /cast or /use.)
--
--  success, reason = TjRunSlash.Execute(command[, slashOnly, [showHelp[, editbox]]])
--    Arguments:
--      command    String. The command to be executed. E.g., "/dance", "/myaddon enable", or "/g Hello, guild!"
--      slashOnly  If evaluates to true, only commands that start with "/" are executed.
--      showHelp   If evaluates to true, text is output to the default chat frame if an unrecognized command is
--                 used. (This is the default text WoW produces when the user enters such a command.)
--      editbox    Table or nil. This should be an edit box frame used for chat input. If nil, it defaults to
--                 ChatFrameEditBox, WoW's standard chat frame edit box. See note below for its purpose.
--    Returns:
--      success    Boolean. True if the command was executed without a problem.
--      reason     String or nil. If a string, gives the reason that the command could not be executed:
--                 "noslash"       The command was not a slash command (and slashOnly was true).
--                 "secure"        A secure command was used.
--                 "unrecognized"  The command was not recognized.
--
--  You can use TjRunSlash(...) as shorthand for TjRunSlash.Execute(...).
--
--  Note: When TjRunSlash.Execute is called, the chat type ("SAY", "GUILD", etc.) it will be called under is set to
--  match that of the given chat edit box (ChatFrameEditBox by default). This allows non-slash command execution to
--  work how most would expect: Using the same chat mode as the player is currently or has been using. Commands that
--  change the chat type (e.g. "/g Hello, guild!") still work as expected.
--

local THIS_VERSION = 0.33

if (not TjRunSlash or TjRunSlash.Version < THIS_VERSION) then
  TjRunSlash = TjRunSlash or {}
  local TjRunSlash = TjRunSlash
  TjRunSlash.Version = THIS_VERSION

  function TjRunSlash.IsSecureCmd(slash, arg)
    assert(type(slash) == "string" and (not arg or type(arg) == "string"),
           "Usage: TjRunSlash.IsSecureCmd( string-slash-command [, string-arguments ] )")
    slash = strlower(slash)
    if (arg and (slash == "/cast" or slash == "/use")) then
    -- Although it is normally considered a secure command, /cast can safely be used with tradeskill "spells":
      if (not TjRunSlash.tradeskills) then
        TjRunSlash.tradeskills = {
          strlower(GetSpellInfo(2259)),  -- Alchemy
          strlower(GetSpellInfo(2018)),  -- Blacksmithing
          strlower(GetSpellInfo(2550)),  -- Cooking
          strlower(GetSpellInfo(7411)),  -- Enchanting
          strlower(GetSpellInfo(4036)),  -- Engineering
          strlower(GetSpellInfo(3273)),  -- First Aid
          strlower(GetSpellInfo(45357)), -- Inscription
          strlower(GetSpellInfo(25229)), -- Jewelcrafting
          strlower(GetSpellInfo(2108)),  -- Leatherworking
          strlower(GetSpellInfo(2842)),  -- Poisons
          strlower(GetSpellInfo(53428)), -- Runeforging
          strlower(GetSpellInfo(2656)),  -- Smelting
          strlower(GetSpellInfo(3908)),  -- Tailoring
        }
      end
      arg = SecureCmdOptionParse(arg)
      if (arg) then
        arg = strtrim( strlower( arg ) )
        for i,v in ipairs(TjRunSlash.tradeskills) do
          if (arg == v) then  return nil;  end
        end
      else
        return nil;
      end
    end
    return IsSecureCmd(slash)
  end

  local fauxChatFrame_AddMessage
  function TjRunSlash.fauxChatFrame()
    if (not fauxChatFrame_AddMessage) then
      fauxChatFrame_AddMessage = function(self, text, ...)
        if (text == HELP_TEXT_SIMPLE) then
          self.UnrecognizedCmd = true;
        end
        if (self.ShowMessages) then
          DEFAULT_CHAT_FRAME:AddMessage(text, ...);
        end
      end
    end
    return { AddMessage = fauxChatFrame_AddMessage }
  end

  local runbox
  function TjRunSlash.Execute(command, slashOnly, showHelp, editbox, extra)
    if (command == TjRunSlash) then  command, slashOnly, showHelp, editbox = slashOnly, showHelp, editbox, extra;  end
    editbox = editbox or ChatFrameEditBox
    if (type(command) ~= "string") then
      error("String expected for first argument (command); got "..type(command)..".")
    elseif (type(editbox) ~= "table") then
      error("Table expected for fifth argument (editbox); got "..type(editbox)..".")
    elseif (slashOnly and strsub(command, 1, 1) ~= "/") then  return false, "noslash";  end
    local slash = strmatch(command, "^(/[^%s]+)")
    if (slash and TjRunSlash.IsSecureCmd(slash, strsub(command, slash:len() + 2))) then  return false, "secure";  end

    if (not runbox) then
      runbox = CreateFrame("EditBox", "TjRunSlash_EditBox", UIParent, "TjRunSlash_EditBoxTemplate")
      -- We needed to create a template in XML for this because using the original template caused its OnLoad
      -- function to run immediately, which caused problems.
    end
    runbox.chatFrame.UnrecognizedCmd = nil
    runbox.chatFrame.ShowMessages = not not showHelp  -- "not not" to make sure it's a boolean
    runbox:SetAttribute("chatType", editbox:GetAttribute("chatType"))
    runbox:SetText(command)
    ChatEdit_SendText(runbox)

    if (runbox.chatFrame.UnrecognizedCmd) then
      return false, "unrecognized"
    end
    return true
  end

  setmetatable(TjRunSlash, { __call = TjRunSlash.Execute })

end
